"""Module which performs cache file related operation stored over CACHE_DIR of OS"""
import importlib
import os
import time
from pathlib import Path
from typing import Optional

import appdirs


def check_if_downloaded(year: int, day: int, session: str) -> bool:
    """Check if an input is downloaded and cached"""
    cache_file = _join_path(year, day, session, file_type="input_file")
    cache_file = Path(cache_file)
    return cache_file.exists()


def save_input_to_location(year: int, day: int, session: str, input_data: str):
    """Save a input to its cache location for future reference and use"""
    cache_folder = _join_path(year, day, session)
    Path(cache_folder).mkdir(parents=True, exist_ok=True)
    cache_file = os.path.join(cache_folder, "input.txt")
    with open(cache_file, "w+") as opened_file:
        opened_file.write(input_data)


def delete_input(year: int, day: int, session: str):
    """Delete input from a cache folder"""
    cache_file = _join_path(year, day, session, file_type="input_file")
    if Path(cache_file).exists():
        os.remove(cache_file)


def cache_file_data(year: int, day: int, session: str) -> str:
    """Return cache file input data from cache folder for certain problem"""
    server_action = importlib.import_module(".server_action")
    server_action.download_input(year, day, session)
    cache_file = _join_path(year, day, session, file_type="input_file")
    with open(cache_file) as opened_file:
        input_data = opened_file.read()
    return input_data


def save_submitted_answer(
    year: int, day: int, part: int, session: str, answer: str, message: str
):
    """Save submitted input to file of problem"""
    submitted_file = _join_path(year, day, session, file_type="submission_file")
    with open(submitted_file, "a") as opened_file:
        opened_file.write("{}!{}:{}\n".format(part, answer, message))


def last_submitted_answer_message(
    year: int, day: int, part: int, session: str, output: str
) -> str:
    """
    Check if answer is already submitted by user if submitted return message of last
    submission
    """
    submission_file = _join_path(year, day, session, file_type="submission_file")
    last_answer_message = ""
    with open(submission_file, "r") as opened_file:
        lines = opened_file.read()
        for line in lines:
            seprate_part = line.split("!", 1)
            if seprate_part[0] == part:
                seprate_output = seprate_part[1].split(":", 1)
                if seprate_output == output:
                    last_answer_message = seprate_output[1]
    return last_answer_message


def save_last_submission_time(year: int, day: int, session: str):
    """Save a time where a request is performed for last submission"""
    last_time_file = _join_path(year, day, session, file_type="last_time_file")
    with open(last_time_file, "w") as opened_file:
        opened_file.write(str(time.time()))


def check_less_than_one_min_submission(year: int, day: int, session: str) -> bool:
    """
    Check last submission time for solution return true if time is less than 60 second
    """
    last_time_file = _join_path(year, day, session, file_type="last_time_file")
    with open(last_time_file, "r") as opened_file:
        last_time = float(opened_file.read())
        current_time = time.time()
        early_submission = current_time - last_time < 60.0
    return early_submission


def _join_path(
    year: int, day: int, session: str, file_type: Optional[str] = None
) -> str:
    """Return desire path for a cache folders or files"""
    cache_location = appdirs.user_cache_dir(appname="advent-of-code")
    cache_file = os.path.join(cache_location, str(session), str(year), str(day))
    if file_type == "input_file":
        cache_file = os.path.join(cache_file, "input.txt")
    if file_type == "submission_file":
        cache_file = os.path.join(cache_file, "submission.txt")
    if file_type == "last_time_file":
        cache_file = os.path.join(cache_file, "time.txt")
    return cache_file



"""module which performs server realted actions such as submitting and downloading"""
import requests

from .cache_file import (
    check_if_downloaded,
    check_less_than_one_min_submission,
    last_submitted_answer_message,
    save_input_to_location,
    save_last_submission_time,
    save_submitted_answer,
)
from .config_file import get_session_value

INPUT_URL = "https://adventofcode.com/{}/day/{}/input"
SUBMIT_URL = "https://adventofcode.com/{}/day/{}/answer"


def download_input(year: int, day: int, session: str):
    """Download file from a advent of code server and save it for future reference"""
    session_value = get_session_value(session)
    if not check_if_downloaded(year, day, session):
        input_url = INPUT_URL.format(year, day)
        html_data = requests.get(input_url, cookies={"session": session_value})
        save_input_to_location(year, day, session, html_data.text)


def submit_output(year: int, day: int, part: int, session: str, output: str) -> str:
    """Submit solution output to a advent of code server"""
    session_value = get_session_value(session)
    submit_url = SUBMIT_URL.format(year, day)
    submitted_message = last_submitted_answer_message(year, day, part, session, output)
    if not submitted_message:
        early_submission = check_less_than_one_min_submission(year, day, session)
        if early_submission:
            message = "You have to wait for 1 min before submitting next solution"
        else:
            data = {"level": part, "answer": output}
            save_last_submission_time(year, day, session)
            response = requests.post(
                submit_url, data, cookies={"session": session_value}
            )
            if response.status_code != 200:
                message = (
                    "Error submiting solution online. Returned non 200 status code"
                )
            else:
                text_data = response.text
                if "too high" in text_data:
                    message = "Your answer is too high"
                    save_submitted_answer(year, day, part, session, output, message)
                elif "too low" in text_data:
                    message = "Your answer is too low"
                    save_submitted_answer(year, day, part, session, output, message)
                elif "That's not" in text_data:
                    message = "That's not the right answer"
                    save_submitted_answer(year, day, part, session, output, message)
                elif "You don't seem" in text_data:
                    message = "You don't seem to be solving right level"
                elif "You gave an answer" in text_data:
                    message = (
                        "You have to wait for 1 min before submitting next solution"
                    )
                elif "That's the right answer" in text_data:
                    message = "Congratulation, you have solved question successfully"
                    save_submitted_answer(
                        year,
                        day,
                        part,
                        session,
                        output,
                        "Congratulation, you have solved question correctly",
                    )
                else:
                    message = (
                        "Got unknown message. Please retry if answered is not submitted"
                    )
        return message
    return submitted_message



"""
advent of code helper module. Provide CLI and library to submit and solve advent of
code problems easily.
"""
__version__ = "0.3.0"

from .initializer import Initializer
from .puzzle import solve, submit
from .runner import advent_runner

__all__ = ["Initializer", "solve", "submit", "advent_runner"]



"""module used for defining puzzle decorator for submiting or solving problem"""
from typing import Callable, List, Optional, TypeVar

from .cache_file import cache_file_data
from .config_file import get_all_session
from .server_action import submit_output

T = TypeVar("T")


class _Puzzle:
    """Puzzle class for handling a puzzle decorator"""

    function: Callable[[str], T]
    year: int
    day: int
    part: int
    session: List[str] = get_all_session()
    operation_type: str
    input_file: Optional[str] = None

    def __repr__(self):
        """Return repr value of class which is set to function __name__"""
        return "{}".format(self.function.__name__)

    def getfunction(self):
        """Get function from a class"""
        return self.function

    def __call__(self):
        """Caller for _Puzzle class"""
        for session_list in self.session:
            if self.input_file is None:
                input_data = cache_file_data(self.year, self.day, session_list)
            else:
                with open(self.input_file) as opened_file:
                    input_data = opened_file.read()
            answer = self.function(input_data)
            if answer is not None:
                if self.operation_type == "submit":
                    message = submit_output(
                        self.year, self.day, self.part, session_list, answer
                    )
                    if message.contains("Congratulation"):
                        print(
                            "{}:{}-{}-{}: {} {} {}".format(
                                session_list,
                                self.year,
                                self.day,
                                self.part,
                                answer,
                                "\u2713",
                                message,
                            )
                        )
                    else:
                        print(
                            "{}:{}-{}-{}: {} {} {}".format(
                                session_list,
                                self.year,
                                self.day,
                                self.part,
                                answer,
                                "\u274C",
                                message,
                            )
                        )
                elif self.operation_type == "solve":
                    print(
                        "{}:{}-{}-{}: {}".format(
                            session_list, self.year, self.day, self.part, answer
                        )
                    )


def submit(
    year: int,
    day: int,
    part: int,
    session_list: Optional[List[str]] = None,
    input_file: Optional[str] = None,
):
    """
    Puzzle decorator used to submit a solution to advent_of_code server and provide
    result. If input_file is not present then it tries to download file and cache it
    for submiting solution else it require to be provided with input_file path which
    input it can use.
    """

    def _action(function):
        operation_type = "submit"
        if not session_list:
            session = get_all_session()
        else:
            session = session_list
        return _Puzzle(
            function=function,
            operation_type=operation_type,
            year=year,
            day=day,
            part=part,
            session=session,
            input_file=input_file,
        )

    return _action


def solve(
    year: int,
    day: int,
    part: int,
    session_list: Optional[List[str]] = None,
    input_file: Optional[str] = None,
):
    """
    Puzzle decorator used to only solve a problem & print output value. It doesn't
    submit output to advent of code server to validate whether an answer is correct
    or not. By default it downloads input file from advent-of-code server. Puzzle
    can also be solved by using custom file location with help of input_file parameter
    while using input_file year, day, part are required even if they are not used for
    only reference purpose of printing output.
    """

    def _action(function):
        operation_type = "solve"
        if not session_list:
            session = get_all_session()
        else:
            session = session_list
        return _Puzzle(
            function=function,
            operation_type=operation_type,
            year=year,
            day=day,
            part=part,
            session=session,
            input_file=input_file,
        )

    return _action



"""Main CLI for advent-of-code helper tool"""
from typing import List

import click

from .cache_file import delete_input
from .config_file import add_to_json, delete_from_json, get_all_session, list_from_json
from .server_action import download_input
from .utils import get_current_year, get_day


@click.group(help="CLI tool to perform action related to advent-of-code")
def main():
    pass


@main.group(help="perform action related to config file")
def config():
    pass


@config.command(help="add session to config")
@click.argument("name")
@click.argument("session_value")
def add(name: str, session_value: str):
    data_list = {name: session_value}
    add_to_json(**data_list)


@config.command("list", help="list all session present in config")
def list_config_session():
    list_from_json()


@config.command("remove", help="remove session from config")
@click.argument("name")
def remove_config_data(name: str):
    delete_from_json(name)


@main.command(help="download solution from advent-of-code server")
@click.option(
    "--year",
    "-y",
    "years",
    multiple=True,
    help="Pass input download year [default: latest year of problem]",
)
@click.option(
    "--day",
    "-d",
    "days",
    multiple=True,
    help="Pass input download day [default: latest problem day or day 1 of old year]",
)
@click.option(
    "--session",
    "-s",
    "sessions",
    multiple=True,
    help="Pass session name or use all session as default",
)
def download(years: List[int], days: List[int], sessions: List[int]):
    if not years:
        years = [get_current_year()]
    if not days:
        days = [get_day()]
    if not sessions:
        sessions = get_all_session()
    for year in years:
        for day in days:
            for session in sessions:
                download_input(year, day, session)


@main.command("remove", help="delete a input file from cache folder")
@click.option(
    "--year",
    "-y",
    "years",
    multiple=True,
    help="Year from which input is to be delete default all",
)
@click.option(
    "--day",
    "-d",
    "days",
    multiple=True,
    help="Day from which input file is to be delete default all day",
)
@click.option(
    "--session",
    "-s",
    "sessions",
    multiple=True,
    help="Session from which input file is to be deleted default all session",
)
def remove_cache(years: List[int], days: List[int], sessions: List[int]):
    if not years:
        years = list(range(2015, get_current_year() + 1))
    if not days:
        days = list(range(1, 26))
    if not sessions:
        sessions = get_all_session()
    for year in years:
        for day in days:
            for session in sessions:
                delete_input(year, day, session)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['advent_of_code_py']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'click>=8.0,<9.0',
 'python-dateutil>=2.8.0,<3.0.0',
 'requests>=2.22.0,<3.0.0']

entry_points = \
{'console_scripts': ['advent-of-code-py = advent_of_code_py.cli:main']}

setup_kwargs = {
    'name': 'advent-of-code-py',
    'version': '0.3.0',
    'description': 'Advent of Code helper CLI and library',
    'long_description': '# Advent-of-code-py\n[Advent of Code][advent_of_code_link] helper CLI and library for python projects.\n\n**Status & Info:**\n\n| Code style | License | Project Version |\n| :---: | :---: | :---: |\n| [![Code style][black_badge]][black_link] | [![License: MIT][license_badge]][license_link] | [![PyPI][project_badge]][project_link] |\n\n## Usage\n\n### Installation\nTo install advent-of-code-py run following command which installs advent-of-code-py CLI and advent_of_code_py library.\n```bash\npip install advent-of-code-py\n```\n\n__OR__\n\n```bash\npoetry add advent-of-code-py\n```\n\n### Usage\nInitially for advent-of-code-py to work it need session value or session ID which you can obtain by viewing cookie while visiting advent of code server.\nAfter collecting session cookie value you need to add those values in config using advent-of-code-py CLI\n```bash\nadvent-of-code-py config add <session-name> <session-value>\n```\n\nNow you can import library by using\n```python\nimport advent_of_code_py\n```\n\nAfter importing a library you can use either two decorator present which are solve and submit decorator for a function of puzzle\n\nFor example:-\n```python\n@advent_of_code_py.submit(2018,3,1,session_list="<session-name>")\ndef puzzle_2018_3_1(data=None):\n    # do some calculation with data and return final output\n    return final_output\n```\n\nNow after decorating function now you can call function like regular function call\n```python\npuzzle_2018_3_1()\n```\nAfter calling function `final_output` value will be submitted by library to Advent of Code server for 2018 year day 3\nproblem, then returns whether the submitted answer was correct or not. If session value is not provided then\nthe solution will be submitted to all session value present in config file.\n\nYou can also use advent-of-code-py builtin Initializer and runner to create appropriate CLI for problem so\nproblem can be run from CLI instead of modifying python file every time to run appropriate function\nTo set advent-of-code-py puzzle as CLI\n```python\n@advent_of_code_py.advent_runner()\ndef main_cli():\n    initializer = advent_of_code_py.Initializer()\n    initializer.add(<function_alias>=<function>)\n    # for example to run above function you can write\n    initializer.add(p_3_1=puzzle_2018_3_1)\n    # add other functions ...\n    return initializer\n```\nNow you can set main_cli as entry points, and it will create CLI with the appropriate name and function which was added.\nSo for example to run function puzzle_2018_3_1() you have to run command as `entry-point-name run p_3_1` which\nwill run the appropriate function as well as submit as desired if the function was decorated by submit decorator or else\nprints its output if the function was decorated by solve decorator.\n\n[advent_of_code_link]: https://adventofcode.com\n\n[black_badge]: https://img.shields.io/badge/code%20style-black-000000.svg?style=for-the-badge\n[black_link]: https://github.com/ambv/black\n\n[license_badge]: https://img.shields.io/github/license/iamsauravsharma/advent-of-code-py.svg?style=for-the-badge\n[license_link]: LICENSE\n\n[project_badge]: https://img.shields.io/pypi/v/advent-of-code-py?style=for-the-badge&color=blue&logo=python\n[project_link]: https://pypi.org/project/advent-of-code-py\n',
    'author': 'Saurav Sharma',
    'author_email': 'appdroiddeveloper@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/iamsauravsharma/advent-of-code-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
initializer module which defines Initializer class for initializing advent of code
function runner
"""
from __future__ import annotations

import inspect
from typing import Callable, Dict, TypeVar

T = TypeVar("T")


class Initializer:
    """
    Initialize class which is used to initialize an advent of code function runner.
    Initialize create a basic class which is used for adding of different
    function of advent of code solution so all of functions can be easily run with
    one simple CLI tool.
    """

    function_list: Dict[str, Callable[[str], T]] = {}

    def add(self, **kwargs):
        """Add a function to a Initializer class"""
        self.function_list.update(kwargs)

    def extend(self, another_initializer: Initializer):
        """Extends initializer with addition of another initialier to it"""
        self.function_list.update(another_initializer.get_function_list())

    def run(self, function_alias: str):
        """Run a certain function by their name/alias"""
        self.function_list.get(function_alias)()

    def run_all(self):
        """Run all functions which are added to `Initializer`"""
        for function in self.function_list.values():
            function()

    def list_functions(self):
        """List all of the function and its alias"""
        for (keys, value) in self.function_list.items():
            print(
                "{:10} -> {}.{}".format(
                    keys, inspect.getmodule(value.getfunction()).__name__, value
                )
            )

    def get_function_list(self):
        """Return all function_list which contain function_alias and function"""
        return self.function_list
